@@ -31,6 +31,9 @@
 
 #ifdef __GNUC__
 #pragma GCC diagnostic ignored "-Waddress"
+#if __GNUC__ >= 6
+# pragma GCC diagnostic ignored "-Wtautological-compare"
+#endif
 #ifdef __clang__
 #  ifdef MZ_PRECISE_GC
 #   pragma clang diagnostic ignored "-Wtautological-compare"
@@ -80,6 +83,11 @@ END_XFORM_ARITH;
 # define _jit_epilog scheme_jit_epilog
 #endif
 
+/* The ABI for _CALL_DARWIN or JIT_X86_64 requires alignment. Even
+   when it's not required, it's better for performance when flonums
+   are stored on the stack. */
+#define JIT_X86_ALIGN_STACK 1
+
 #ifndef DEFINE_LIGHTNING_FUNCS
 # define SUPPRESS_LIGHTNING_FUNCS
 #endif
@@ -294,7 +302,6 @@ struct scheme_jit_common_record {
 
   void *bad_result_arity_code;
   void *unbound_global_code;
-  void *quote_syntax_code;
   void *call_original_unary_arith_code;
   void *call_original_binary_arith_code;
   void *call_original_binary_rev_arith_code;
@@ -307,24 +314,25 @@ struct scheme_jit_common_record {
   void *bad_cXr_code;
   void *bad_mcar_code, *bad_mcdr_code;
   void *bad_set_mcar_code, *bad_set_mcdr_code;
+  void *bad_syntax_e_code;
   void *imag_part_code, *real_part_code, *make_rectangular_code;
   void *bad_flimag_part_code, *bad_flreal_part_code, *bad_make_flrectangular_code;
-  void *unbox_code, *set_box_code, *box_cas_fail_code;
+  void *unbox_code, *set_box_code, *unbox_star_fail_code, *set_box_star_fail_code, *box_cas_fail_code, *weak_box_value_code;
   void *vector_cas_fail_code;
-  void *bad_vector_length_code;
+  void *bad_vector_length_code, *bad_vector_star_length_code;
   void *bad_flvector_length_code;
   void *bad_fxvector_length_code;
   void *bad_string_length_code;
   void *bad_bytes_length_code;
   void *vector_ref_code, *vector_ref_check_index_code, *vector_set_code, *vector_set_check_index_code;
   void *chap_vector_ref_code, *chap_vector_ref_check_index_code, *chap_vector_set_code, *chap_vector_set_check_index_code;
+  void *vector_star_ref_code, *vector_star_ref_check_index_code, *vector_star_set_code, *vector_star_set_check_index_code;
   void *string_ref_code, *string_ref_check_index_code, *string_set_code, *string_set_check_index_code;
   void *bytes_ref_code, *bytes_ref_check_index_code, *bytes_set_code, *bytes_set_check_index_code;
   void *flvector_ref_check_index_code[JIT_NUM_FL_KINDS];
   void *flvector_set_check_index_code[JIT_NUM_FL_KINDS], *flvector_set_flonum_check_index_code[JIT_NUM_FL_KINDS];
   void *fxvector_ref_code, *fxvector_ref_check_index_code, *fxvector_set_code, *fxvector_set_check_index_code;
   void *struct_raw_ref_code, *struct_raw_set_code, *struct_raw_refs_code;
-  void *syntax_e_code;
   void *on_demand_jit_arity_code, *in_progress_on_demand_jit_arity_code;
   void *get_stack_pointer_code;
   void *stack_cache_pop_code;
@@ -344,11 +352,13 @@ struct scheme_jit_common_record {
   void *bad_char_to_integer_code, *slow_integer_to_char_code;
   void *slow_cpointer_tag_code, *slow_set_cpointer_tag_code;
   void *values_code;
+  void *symbol_interned_p_code;
   void *list_p_code, *list_p_branch_code;
   void *list_length_code;
   void *list_ref_code, *list_tail_code;
+  void *hash_ref_code;
   void *finish_tail_call_code, *finish_tail_call_fixup_code;
-  void *module_run_start_code, *module_exprun_start_code, *module_start_start_code;
+  void *linklet_run_start_code;
   void *thread_start_child_code;
   void *box_flonum_from_stack_code, *box_flonum_from_reg_code;
   void *fl1_fail_code[JIT_NUM_FL_KINDS], *fl2rr_fail_code[2][JIT_NUM_FL_KINDS];
@@ -546,7 +556,7 @@ void *scheme_jit_get_threadlocal_table();
 #  ifdef THREAD_LOCAL_USES_JIT_V2
 #   define mz_tl_addr(reg, addr) (jit_addi_p(reg, JIT_V2, addr))
 #   define mz_tl_addr_tmp(tmp_reg, addr) (void)0
-#   define mz_tl_addr_untmp(tmp_reg) 0
+#   define mz_tl_addr_untmp(tmp_reg) (void)0
 #   define mz_tl_tmp_reg(tmp_reg) (void)0
 #   define _mz_tl_str_p(addr, tmp_reg, reg) jit_stxi_p(addr, JIT_V2, reg)
 #   define _mz_tl_str_l(addr, tmp_reg, reg) jit_stxi_l(addr, JIT_V2, reg)
@@ -818,7 +828,7 @@ static void *top;
       LOCAL1 (which is a cont_mark_stack offset, if anything)
       LOCAL2 (some pointer, never to stack or runstack)
       LOCAL3 (temp space for misc uses; not saved across calls that might capture LWC)
-      LOCAL4 (x86_64: = saved R14 otherwise when THREAD_LOCAL
+      LOCAL4 (x86_64: = saved R14 when THREAD_LOCAL
               x86: = RUNSTACK_BASE or THREAD_LOCAL)
       [some empty slots, maybe, depending on alignment]
       [space for "flostack" --- local unboxed values, such as flonums]
@@ -931,20 +941,13 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
 # define mz_get_local_p_x(x, l, FP) jit_ldxi_p((x), FP, (l))
 # define mz_patch_branch_at(a, v) jit_patch_branch_at(a, v)
 # define mz_patch_ucbranch_at(a, v) jit_patch_ucbranch_at(a, v)
-  /* The ABI for _CALL_DARWIN or JIT_X86_64 requires alignment. Even
-     when it's not required, it's better for performance when flonums
-     are stored on the stack. */
-# define X86_ALIGN_STACK 1
-# ifdef X86_ALIGN_STACK
+# ifdef JIT_X86_ALIGN_STACK
    /* Maintain 16-byte stack alignment. */
 #  ifdef JIT_X86_64
 #   define STACK_ALIGN_WORDS 1
 #  else
 #   define STACK_ALIGN_WORDS 3
 #  endif
-#  define mz_prolog(x) (ADDQiBr(-(STACK_ALIGN_WORDS * JIT_WORD_SIZE), JIT_SP))
-#  define mz_epilog_without_jmp() ADDQiBr((STACK_ALIGN_WORDS + 1) * JIT_WORD_SIZE, JIT_SP)
-#  define mz_epilog(x) (ADDQiBr(STACK_ALIGN_WORDS * JIT_WORD_SIZE, JIT_SP), RET_())
 #  define JIT_LOCAL3 -(JIT_WORD_SIZE * 6)
 #  ifdef NEED_LOCAL4
 #   ifdef JIT_X86_64
@@ -956,10 +959,10 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
 #  else
 #   define LOCAL_FRAME_SIZE 3
 #  endif
+#  define _mz_prolog(x) (ADDQiBr(-(STACK_ALIGN_WORDS * JIT_WORD_SIZE), JIT_SP))
+#  define _mz_epilog_without_jmp() ADDQiBr((STACK_ALIGN_WORDS + 1) * JIT_WORD_SIZE, JIT_SP)
+#  define _mz_epilog(x) (ADDQiBr(STACK_ALIGN_WORDS * JIT_WORD_SIZE, JIT_SP), RET_())
 # else
-#  define mz_prolog(x) /* empty */
-#  define mz_epilog(x) RET_()
-#  define mz_epilog_without_jmp() ADDQir(JIT_WORD_SIZE, JIT_SP)
 #  define JIT_LOCAL3 JIT_LOCAL2
 #  ifdef NEED_LOCAL4
 #   define LOCAL_FRAME_SIZE 3
@@ -967,10 +970,29 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
 #  else
 #   define LOCAL_FRAME_SIZE 2
 #  endif
+#  define _mz_prolog(x) /* empty */
+#  define _mz_epilog(x) RET_()
+#  define _mz_epilog_without_jmp() ADDQir(JIT_WORD_SIZE, JIT_SP)
 # endif
 # ifdef NEED_LOCAL4
 #   define JIT_LOCAL4 -(JIT_WORD_SIZE * JIT_LOCAL4_OFFSET)
 # endif
+# ifdef MZ_PROLOG_CREATE_FULL_STACK_FRAME
+  /* Make the internal ABI the same as the main call ABI */
+#  define MZ_LOCAL_FRAME_SIZE (LOCAL_FRAME_SIZE+3)
+#  define mz_prolog(x) (PUSHQr(_EBP),                                  \
+                        mz_get_local_p((x), JIT_LOCAL3),               \
+                        MOVQrr(_ESP, _EBP),                            \
+                        ADDQiBr(-(MZ_LOCAL_FRAME_SIZE * JIT_WORD_SIZE), JIT_SP), \
+                        mz_set_local_p((x), JIT_LOCAL3))
+#  define mz_epilog_without_jmp() (ADDQiBr(MZ_LOCAL_FRAME_SIZE * JIT_WORD_SIZE, JIT_SP), POPQr(_EBP), ADDQiBr(JIT_WORD_SIZE, JIT_SP))
+#  define mz_epilog(x) (ADDQiBr(MZ_LOCAL_FRAME_SIZE * JIT_WORD_SIZE, JIT_SP), POPQr(_EBP), RET_())
+# else
+  /* Normal internal ABI */
+#  define mz_prolog(x) _mz_prolog(x)
+#  define mz_epilog_without_jmp() _mz_epilog_without_jmp()
+#  define mz_epilog(x) _mz_epilog(x)
+# endif
 # define mz_push_locals() SUBQir((LOCAL_FRAME_SIZE << JIT_LOG_WORD_SIZE), JIT_SP)
 # define mz_pop_locals() ADDQir((LOCAL_FRAME_SIZE << JIT_LOG_WORD_SIZE), JIT_SP)
 # define JIT_FRAME_FLOSTACK_OFFSET (-(JIT_WORD_SIZE * (LOCAL_FRAME_SIZE + 3)))
@@ -1420,7 +1442,7 @@ int scheme_generate_inlined_test(mz_jit_state *jitter, Scheme_Object *obj, int b
                                  Branch_Info *for_branch);
 int scheme_generate_cons_alloc(mz_jit_state *jitter, int rev, int inline_retry, int known_list, int dest);
 int scheme_generate_struct_alloc(mz_jit_state *jitter, int num_args, 
-                                 int inline_slow, int pop_and_jump,
+                                 int inline_slow, int pop_and_jump, int check_proc, 
                                  int is_tail, int multi_ok, int dest);
 int scheme_generate_two_args(Scheme_Object *rand1, Scheme_Object *rand2, mz_jit_state *jitter, 
                              int order_matters, int skipped);
@@ -1456,6 +1478,7 @@ int scheme_generate_unboxing(mz_jit_state *jitter, int target);
 int scheme_generate_pop_unboxed(mz_jit_state *jitter);
 int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
                                int arith, int cmp, Branch_Info *for_branch, int branch_short,
+                               int unsafe_fx, int unsafe_fl,
                                int dest);
 int scheme_generate_alloc_double(mz_jit_state *jitter, int inline_retry, int dest);
 int scheme_generate_arith(mz_jit_state *jitter, Scheme_Object *rator, Scheme_Object *rand, Scheme_Object *rand2, 
@@ -1529,7 +1552,7 @@ int scheme_generate_struct_op(mz_jit_state *jitter, int kind, int for_branch,
                               int result_ignored,
                               int check_proc, int check_arg_fixnum,
                               int type_pos, int field_pos,
-                              int authentic,
+                              int authentic, int type_unpacked,
                               int pop_and_jump,
                               jit_insn *refslow, jit_insn *refslow2,
                               jit_insn *bref_false, jit_insn *bref_true);
@@ -1594,7 +1617,7 @@ int scheme_jit_check_closure_extflonum_bit(Scheme_Lambda *data, int pos, int del
 
 Scheme_Object *scheme_extract_global(Scheme_Object *o, Scheme_Native_Closure *nc, int local_only);
 Scheme_Object *scheme_extract_closure_local(Scheme_Object *obj, mz_jit_state *jitter, int extra_push, int get_constant);
-Scheme_Object *scheme_specialize_to_constant(Scheme_Object *obj, mz_jit_state *jitter, int extra_push);
+Scheme_Object *scheme_specialize_to_constant(Scheme_Object *obj, mz_jit_state *jitter, int extra_push, int extract_static);
 
 void scheme_jit_register_traversers(void);
 #ifdef MZ_USE_LWC
